/* CallWin32Podule.cpp
Provides the podule's functionality
Original by Graeme Barnes: My changes (C) 2002 Niall Douglas
8th August 2002 (started: 16th July 2002)

My changes are provided under the GNU General Public Licence.
See the file gpl.txt in the archive for more information.
*/

#include "stdafx.h"
#include "resource.h"

// unfortunate clash...
#undef LoadString

#include "Console.h"
#include "UserInterface.h"
#include "Messages.h"
#include "CallWin32Podule.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>



// register map
#define REG_PAGEROM				0x2000
#define REG_PAGERAM				0x2004
// Read to get interrupt status, write clears it
#define REG_INTSTATUS			0x2008
// Read to determine source of interrupt
#define REG_VECTOR				0x200c
// Write no of pars to reset for new call
#define REG_RESET				0x2010
// Read to call DLL and obtain result
#define REG_CALL				0x2014
// Read to obtain error code
#define REG_ERRORCODE			0x2018
// Read to obtain Win32 error code
#define REG_WIN32ERROR			0x201c
// Parameter no=0 for function name, =1 for first par etc etc
#define REG_PARNO				0x2020
// A value
#define REG_PARVALUE			0x2024
// Latches parameter data into the device
#define REG_LATCHDATA			0x2028
#define REG_PARDATA				0x202c

// rom is accessed on word boundaries and spread into 0x2000 bytes
// ie half of the module workspace
#define ROM_PAGE_SHIFT 11
#define ROM_PAGE_SIZE (1<<ROM_PAGE_SHIFT)
#define ROM_PAGE_MASK (ROM_PAGE_SIZE-1)

// Exported C-style function used by other DLL's to cause a vector call
static CCallWinPodule *pod;
extern "C" void __declspec(dllexport) __cdecl CallWinPodule_SignalVector(int vector)
{
	if(pod) pod->CallVector(vector);
}
// Exported C-style function used by other DLL's to fetch the RO window handle
static HWND _parent, _riscos;
static struct UpcallRoutine
{
	void (_cdecl *code)(int code, void *data);
	void *data;
} upcalls[16];
extern "C" void __declspec(dllexport) __cdecl CallWinPodule_GetRISCOSWindowHandles(HWND *parent, HWND *riscos)
{
	if(pod)
	{
		if(!_parent)
		{
			IWinUI *ui = (IWinUI*) pod->FindInterface(IFACE_USERINTERFACE);
			_parent=ui->GetAppWindowHandle(ui);
			_riscos=ui->GetClientWindowHandle(ui);
		}
		if(parent) *parent=_parent;
		if(riscos) *riscos=_riscos;
	}
}
// Exported C-style function used by other DLL's (probably only MouseSync) to disable mouse input
extern "C" void __declspec(dllexport) __cdecl CallWinPodule_SetMouseInputDisable(int mask)
{
	if(pod)
	{
		IWinUI *ui = (IWinUI*) pod->FindInterface(IFACE_USERINTERFACE);
		ui->EnableMouseInput(ui, mask);
	}
}
// Exported C-style function used by other DLL's (probably only MouseSync) to know when full screen state has changed
extern "C" int __declspec(dllexport) _cdecl CallWinPodule_InstallUpcall(void (_cdecl *code)(int code, void *data), void *data)
{
	for(int n=0; n<16; n++)
	{
		if(!upcalls[n].code) break;
	}
	if(16==n) return false;
	upcalls[n].code=code;
	upcalls[n].data=data;
	return true;
}
//Exported C-style function used by RISC-OS to determine the RS version
extern "C" void __declspec(dllexport) _cdecl CallWinPodule_GetRedSquirrelVersion(int *versions, char *string)
{	// For 0.6.0
	Word d=CComponentManager::GetAppVersion();
	versions[0]=(d>>24) & 0xff;
	versions[1]=(d>>16) & 0xff;
	versions[2]=(d>>8) & 0xff;
	versions[3]=(d) & 0xff;
	strcpy(string, (versions[1]) ? "VirtualAcorn" : "Red Squirrel");
	sprintf(strchr(string,0), " v%d.%d", versions[2], versions[3]);
}

// declare the component so it can be created from a model file
IMPLEMENT_COMPONENT("CallWin32Podule", CCallWinPodule);

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


// Constructor. Only basic initialisation should be done here
CCallWinPodule::CCallWinPodule()
{
	// Only simple initialisation in here
	rombase = 0;
	romsize = 0;
	rompage = 0;
	int_status=0;
	int_mask=0;
	vector=0;

	RAMoffset=0;
}

CCallWinPodule::~CCallWinPodule()
{

}

Word CCallWinPodule::Ioctl(Word op, void* in, void* out)
{
	if(MESSAGEMAJOR(op)==MESSAGE_BROADCAST)
	{
		DisplaySurfaceInfo *si=(DisplaySurfaceInfo *) in;
		if(MESSAGEMINOR(op)==MB_DISPLAY_SURFACE_CHANGED)
		{
			bool fullScreen=!((_parent==si->parent) && (_riscos==si->child));
			for(int n=0; n<16; n++)
			{
				if(upcalls[n].code)
					(upcalls[n].code)(fullScreen, upcalls[n].data);
			}
		}
	}
	return CIocPodule::Ioctl(op, in, out);
}

// In init we perform local tasks such as allocating memory
BOOL CCallWinPodule::Init()
{
	// Load the rom
	const char* romname = LoadString("rom", "callwin32.dat");
	char path[1024];

	FILE* fp = 0;
	for (const char** spath = CComponentManager::GetSearchPath(); *spath; ++spath)
	{
		strcpy(path, *spath);
		strcat(path, romname);

		fp = fopen(path, "rb");
		if (fp)
			break;
	}
	if (fp)
	{
		fseek(fp, 0, SEEK_END);
		romsize = ftell(fp);
		fseek(fp, 0, SEEK_SET);

		rombase = (Byte*)malloc(romsize);
		fread(rombase, 1, romsize, fp);
		fclose(fp);
	}
	else
	{
		console->printf("CallWinPodule failed to load rom image\n");
		return FALSE;
	}

	// Point to ourselves
	pod=this;

	return CIocPodule::Init();
}

// In connect we connect to other components
BOOL CCallWinPodule::Connect()
{
	return CIocPodule::Connect();
}

BOOL CCallWinPodule::Disconnect()
{
	return CIocPodule::Disconnect();
}

BOOL CCallWinPodule::Cleanup()
{
	free(rombase);
	rombase = 0;
	romsize = 0;
	rompage = 0;

	pod=0;
	
	return CIocPodule::Cleanup();
}

// A timer event happend. Just signal an interrupt and carry on
void CCallWinPodule::CallVector(Word v)
{
	vector=v;
	int_status=int_mask;
	if(int_status) SetIrq();
}

void CCallWinPodule::Write(Word address, Word w, Byte width)
{
	if(REG_PAGEROM==address)
	{
		if ((w<<ROM_PAGE_SHIFT) < romsize)
			rompage = w;
		else
			rompage = 0;
	}
	else if(REG_PAGERAM==address)
	{
		RAMoffset=w;
	}
	else if(REG_INTSTATUS==address)
	{
		int_status = 0;
		int_mask = w;
		ClearIrq();
	}
	else if(REG_RESET==address)
	{
		DLLname[0]=0;
		APIname[0]=0;
		errorcode=0;
		lasterror=0;
		currentpar=0;
		totalpars=w;
		memset(pars, 0, sizeof(pars));
		for(int i=0; i<PARS_MAX; i++) pardata[i].idx=0;
	}
	else if(REG_CALL==address)
	{
		// Do nothing on write
	}
	else if(REG_PARNO==address)
	{
		currentpar=w;
	}
	else if(REG_PARVALUE==address)
	{
		pars[currentpar]=w;
	}
	else if(REG_LATCHDATA==address)
	{
		pars[currentpar]=(Word) pardata[currentpar].data;
	}
	else if(address>=REG_PARDATA)
	{
		*(Word *)(&pardata[currentpar].data[pardata[currentpar].idx])=w;
		pardata[currentpar].idx+=sizeof(Word);
		if(currentpar>PARDATA_MAX/sizeof(char)) currentpar=0;
	}
}

// handle word reads from our podule
Word CCallWinPodule::Read(Word address, Byte width)
{
	if(REG_INTSTATUS==address)
		return int_status;
	else if(REG_VECTOR==address)
		return vector;
	else if(REG_CALL==address)
	{
		int callconv=0;
		char *desc=pardata[0].data;
		char *sep1=strchr(desc, ':'), *sep2=strstr(desc, "::");
		if(!sep1 || !sep2)
		{
			errorcode=1;	// Bad API descript
			return 0;
		}
		callconv=desc[0];
		if(callconv!='s' && callconv!='c')
		{
			errorcode=1;	// Bad API descript
			return 0;
		}
		memcpy(DLLname, sep1+1, sep2-sep1-1);
		DLLname[sep2-sep1-1]=0;
		strcpy(APIname, sep2+2);

		HANDLE h=GetModuleHandle(DLLname);
		if(!h) h=LoadLibrary(DLLname);
		if(!h)
		{
			errorcode=2;	// DLL not found
			return 0;
		}
		void *funct=GetProcAddress((HINSTANCE) h, APIname);
		if(!funct)
		{
			errorcode=3;	// API not found
			return 0;
		}
		Word retval, _totalpars=totalpars*sizeof(Word);
		for(int i=totalpars; i>0; i--)
		{
			Word foo=pars[i];
			__asm
			{
				push foo
			}
		}
		if('s'==callconv)		// stdcall
			__asm
			{
				call funct
				mov retval, eax
			}
		else if('c'==callconv)	// cdecl
			__asm
			{
				call funct
				add esp, _totalpars
				mov retval, eax
			}
		lasterror=GetLastError();
		for(int p=0; p<PARS_MAX; p++) pardata[p].idx=0;
		return retval;
	}
	else if(REG_ERRORCODE==address)
		return errorcode;
	else if(REG_WIN32ERROR==address)
		return lasterror;
	else if(address>=REG_PARDATA)
	{
		Word ret=*(Word *)(&pardata[currentpar].data[pardata[currentpar].idx]);
		pardata[currentpar].idx+=sizeof(Word);
		if(currentpar>PARDATA_MAX/sizeof(char)) currentpar=0;
		return ret;
	}
	else if(rombase)
	{
		Word romaddr = ((address>>2)&ROM_PAGE_MASK) + (rompage<<ROM_PAGE_SHIFT);
		return rombase[romaddr];
	}
	else
		return 0xffffffff;
}

#if 0
// stdcall unoptimised
mov esi,esp
mov edx, dword ptr [ebp-4]
mov eax, dword ptr [edx+12Ch]
push eax
mov edx, dword ptr [ebp-4]
mov eax, dword ptr [edx+128h]
push eax
...
call dword ptr [cfunct]
cmp esi,esp
call _chkesp
mov dword ptr [ret],eax

// stdcall optimised
// eax is return from GetProcAddress. esi is returned in eax at the end.
push dword ptr [ebx+12Ch]
push dword ptr [ebx+128h]
push dword ptr [ebx+124h]
push dword ptr [ebx+120h]
push dword ptr [ebx+11Ch]
push dword ptr [ebx+118h]
call eax
mov esi,eax

// cdecl optimised
push dword ptr [ebx+12Ch]
push dword ptr [ebx+128h]
push dword ptr [ebx+124h]
push dword ptr [ebx+120h]
push dword ptr [ebx+11Ch]
push dword ptr [ebx+118h]
call eax
add esp,18h
mov esi,eax

// fastcall optimised
push dword ptr [ebx+12Ch]
mov edx, dword ptr [ebx+11Ch]
mov ecx, dword ptr [ebx+118h]
push dword ptr [ebx+128h]
push dword ptr [ebx+124h]
push dword ptr [ebx+120h]
call eax
mov esi,eax

#endif